/**
   @author Shin'ichiro Nakaoka
*/

#include "ItemTreeView.h"
#include "Item.h"
#include "RootItem.h"
#include "ExtensionManager.h"
#include "ItemManager.h"
#include "MenuManager.h"
#include "AppConfig.h"
#include "Archive.h"
#include "TreeWidget.h"
#include <cnoid/ConnectionSet>
#include <QBoxLayout>
#include <QApplication>
#include <QMouseEvent>
#include <QHeaderView>
#include <boost/function.hpp>
#include <boost/bind.hpp>
#include <set>
#include <cassert>
#include <iostream>
#include "gettext.h"

using namespace std;
using namespace boost;
using namespace cnoid;

namespace {

    // Item of the item tree view
    class ItvItem : public QTreeWidgetItem
    {
    public:
        ItvItem(Item* item, ItemTreeViewImpl* itemTreeViewImpl);
        virtual ~ItvItem();
        virtual QVariant data(int column, int role) const;
        virtual void setData(int column, int role, const QVariant& value);
        ItemPtr item;
        boost::signal<void(bool isChecked)> sigCheckToggled;
        ItemTreeViewImpl* itemTreeViewImpl;
        bool isExpandedBeforeRemoving;
    };

    // Preserved as custom data in an item
    class ItvItemRef : public Referenced
    {
    public:
        ItvItemRef(ItvItem* itvItem) : itvItem(itvItem) { }
        ItvItem* itvItem;
    };
}


namespace cnoid {

    class ItemTreeViewImpl : public TreeWidget, public boost::signals::trackable
    {
    public:

        ItemTreeViewImpl(ItemTreeView* self, RootItem* rootItem, bool showRoot);
        ~ItemTreeViewImpl();

        ItemTreeView* self;
        RootItemPtr rootItem;

        int isProceccingSlotForRootItemSignals;
        ConnectionSet connectionsFromRootItem;

        boost::signal<void(Item* item, bool isChecked)> sigCheckToggled;
        boost::signal<void(bool isChecked)> sigCheckToggledForInvalidItem;
        boost::signal<void(const ItemList<>&)> sigSelectionChanged;
        boost::signal<void(const ItemList<>&)> sigSelectionOrTreeChanged;

        ItemList<> selectedItemList;
        ItemList<Item> copiedItemList;
        ItemList<> checkedItemList;
        bool needToUpdateCheckedItemList;
        Menu* popupMenu;
        MenuManager menuManager;

        bool isDropping;
        int fontPointSizeDiff;

        virtual void mousePressEvent(QMouseEvent* event);
        virtual void keyPressEvent(QKeyEvent* event);
        
        ItvItem* getItvItem(Item* item);
        ItvItem* getOrCreateItvItem(Item* item);
        void onSubTreeAddedOrMoved(Item* item);
        void insertItem(QTreeWidgetItem* parentTwItem, Item* item, Item* nextItem);
        void onSubTreeRemoved(Item* item, bool isMoving);
        void onTreeChanged();
        void onItemAssigned(Item* assigned, Item* srcItem);

        virtual void dropEvent(QDropEvent* event);
        
        void onRowsAboutToBeRemoved(const QModelIndex& parent, int start, int end);
        void onRowsInserted(const QModelIndex& parent, int start, int end);
        virtual bool dropMimeData(QTreeWidgetItem* parent, int index, const QMimeData* data, Qt::DropAction action);
        void onSelectionChanged();
        bool isItemSelected(Item* item);
        bool selectItem(Item* item, bool select);
        bool isItemChecked(Item* item);
        bool checkItem(Item* item, bool checked);
        void extractSelectedItemsOfSubTreeTraverse(Item* item, ItemList<>* io_items);
        ItemList<>& checkedItems();
        void extractCheckedItems(QTreeWidgetItem* twItem);
        void forEachTopItems(const ItemList<>& orgItemList, function<void(Item*)> callback);
        void cutSelectedItems();
        void copySelectedItems();
        void copySelectedItemsSub(Item* item, ItemPtr& duplicated, set<Item*>& items);
        void copySelectedItemsWithChildren();
        void addCopiedItemToCopiedItemList(Item* item);
        void pasteItems();
        void moveCutItemsToCopiedItemList(Item* item);
        void checkSelectedItems(bool on);
        void toggleSelectedItemChecks();
        bool storeState(Archive& archive);
        void storeItemIds(Archive& archive, const char* key, const ItemList<>& items);
        bool restoreState(const Archive& archive);
        void restoreItemStates(const Archive& archive, const char* key, function<void(ItemPtr)> stateChangeFunc);
        void storeExpandedItems(Archive& archive);
        void storeExpandedItemsSub(QTreeWidgetItem* parentTwItem, Archive& archive, ListingPtr& expanded);
        void restoreExpandedItems(const Archive& archive);
        void zoomFontSize(int pointSizeDiff);
    };
}


ItvItem::ItvItem(Item* item, ItemTreeViewImpl* itemTreeViewImpl)
    : item(item),
      itemTreeViewImpl(itemTreeViewImpl)
{
    
    setFlags(Qt::ItemIsEnabled | Qt::ItemIsSelectable | Qt::ItemIsUserCheckable);

    if(!item->isSubItem()){
        setFlags(flags() | Qt::ItemIsEditable | Qt::ItemIsDragEnabled | Qt::ItemIsDropEnabled);
    }

    setCheckState(1, Qt::Unchecked);
    
    ItvItemRef* ref = new ItvItemRef(this);
    item->setCustomData(0, ref);

    isExpandedBeforeRemoving = false;
}


ItvItem::~ItvItem()
{
    item->clearCustomData(0);
}


QVariant ItvItem::data(int column, int role) const
{
    if((role == Qt::DisplayRole || role == Qt::EditRole) && column == 0){
        return item->name().c_str();
    } else if(column == 1 && (role == Qt::CheckStateRole)) {
        return QTreeWidgetItem::data(column, role);
    } else {
        return QVariant();
    }
}


void ItvItem::setData(int column, int role, const QVariant& value)
{
    bool emitCheckToggled = false;
    
    if(column == 0){
        if(role == Qt::DisplayRole || role == Qt::EditRole){
            if(value.type() == QVariant::String){
                if(!value.toString().isEmpty()){
                    item->setName(value.toString().toStdString());
                    QTreeWidgetItem::setData(column, role, value);
                }
            }
        }
    } else if(column == 1 && role == Qt::CheckStateRole){
        emitCheckToggled = true;
        QTreeWidgetItem::setData(column, role, value);
    }
    
    if(emitCheckToggled){
        itemTreeViewImpl->needToUpdateCheckedItemList = true;
        bool checked = ((Qt::CheckState)value.toInt() == Qt::Checked);
        itemTreeViewImpl->sigCheckToggled(item.get(), checked);
        sigCheckToggled(checked);
    }
}


void ItemTreeView::initialize(ExtensionManager* ext)
{
    static bool initialized = false;
    if(!initialized){
        ext->addView(mainInstance());
        initialized = true;
    }
}


ItemTreeView* ItemTreeView::instance()
{
    static ItemTreeView* itemTreeView = new ItemTreeView(RootItem::mainInstance());
    return itemTreeView;
}


ItemTreeView* ItemTreeView::mainInstance()
{
    return instance();
}


ItemTreeView::ItemTreeView(RootItem* rootItem, bool showRoot)
{
    setName(N_("Items"));
    setDefaultLayoutArea(View::LEFT);
    
    impl = new ItemTreeViewImpl(this, rootItem, showRoot);

    QVBoxLayout* layout = new QVBoxLayout();
    layout->addWidget(impl);
    setLayout(layout);
}


ItemTreeViewImpl::ItemTreeViewImpl(ItemTreeView* self, RootItem* rootItem, bool showRoot)
    : self(self),
      rootItem(rootItem)
{
    isProceccingSlotForRootItemSignals = 0;
    needToUpdateCheckedItemList = false;
    isDropping = false;
    
    setColumnCount(2);
    header()->setStretchLastSection(false);
#if QT_VERSION < QT_VERSION_CHECK(5, 0, 0)
    header()->setResizeMode(0, QHeaderView::Stretch);
    header()->setResizeMode(1, QHeaderView::ResizeToContents);
    header()->setMinimumSectionSize(0);
#else
    header()->setSectionResizeMode(0, QHeaderView::Stretch);
    header()->setSectionResizeMode(1, QHeaderView::ResizeToContents);
    //header()->setMinimumSectionSize(0);
#endif
    header()->swapSections(0, 1);
    setWordWrap(true);
    setFrameShape(QFrame::NoFrame);
    setHeaderHidden(true);
    setIndentation(12);
    setSelectionMode(QAbstractItemView::ExtendedSelection);
    setDragDropMode(QAbstractItemView::InternalMove);

    sigItemSelectionChanged().connect(bind(&ItemTreeViewImpl::onSelectionChanged, this));

    connectionsFromRootItem.add(
        rootItem->sigSubTreeAdded().connect(bind(&ItemTreeViewImpl::onSubTreeAddedOrMoved, this, _1)));
    connectionsFromRootItem.add(
        rootItem->sigSubTreeMoved().connect(bind(&ItemTreeViewImpl::onSubTreeAddedOrMoved, this, _1)));
    connectionsFromRootItem.add(
        rootItem->sigSubTreeRemoved().connect(bind(&ItemTreeViewImpl::onSubTreeRemoved, this, _1, _2)));
    connectionsFromRootItem.add(
        rootItem->sigTreeChanged().connect(bind(&ItemTreeViewImpl::onTreeChanged, this)));
    connectionsFromRootItem.add(
        rootItem->sigItemAssigned().connect(bind(&ItemTreeViewImpl::onItemAssigned, this, _1, _2)));

    QObject::connect(model(), SIGNAL(rowsAboutToBeRemoved(const QModelIndex&, int, int)),
                     self, SLOT(onRowsAboutToBeRemoved(const QModelIndex&, int, int)));
    QObject::connect(model(), SIGNAL(rowsInserted(const QModelIndex&, int, int)),
                     self, SLOT(onRowsInserted(const QModelIndex&, int, int)));

    popupMenu = new Menu(this);
    menuManager.setTopMenu(popupMenu);

    menuManager.addItem(_("Cut"))
        ->sigTriggered().connect(bind(&ItemTreeViewImpl::cutSelectedItems, this));
    menuManager.addItem(_("Copy"))
        ->sigTriggered().connect(bind(&ItemTreeViewImpl::copySelectedItems, this));
    menuManager.addItem(_("Copy with Children"))
        ->sigTriggered().connect(bind(&ItemTreeViewImpl::copySelectedItemsWithChildren, this));
    menuManager.addItem(_("Paste"))
        ->sigTriggered().connect(bind(&ItemTreeViewImpl::pasteItems, this));

    menuManager.addSeparator();
    menuManager.addItem(_("Check"))
        ->sigTriggered().connect(bind(&ItemTreeViewImpl::checkSelectedItems, this, true));
    menuManager.addItem(_("Uncheck"))
        ->sigTriggered().connect(bind(&ItemTreeViewImpl::checkSelectedItems, this, false));
    menuManager.addItem(_("Toggle checks"))
        ->sigTriggered().connect(bind(&ItemTreeViewImpl::toggleSelectedItemChecks, this));

    menuManager.addSeparator();
    menuManager.addItem(_("Select all"))
        ->sigTriggered().connect(bind(&ItemTreeView::selectAllItems, self));
    menuManager.addItem(_("Clear selection"))
        ->sigTriggered().connect(bind(&ItemTreeView::clearSelection, self));
    
    fontPointSizeDiff = 0;
    MappingPtr config = AppConfig::archive()->openMapping("ItemTreeView");
    int storedFontPointSizeDiff;
    if(config->read("fontZoom", storedFontPointSizeDiff)){
        zoomFontSize(storedFontPointSizeDiff);
    }
}


ItemTreeView::~ItemTreeView()
{

}


ItemTreeViewImpl::~ItemTreeViewImpl()
{
    // On Windows + VC++, boost::signal::trackable, the super class of
    // ItemTreeViewImpl, does not seem to work correctly, and the connection
    // is not disconnected and the program aborted by segmentation fault.
    // So the following explicit disconnection code is added.

    //connectionClassUnregistered.disconnect();
}


RootItem* ItemTreeView::rootItem()
{
    return impl->rootItem.get();
}


void ItemTreeView::showRoot(bool show)
{

}


void ItemTreeViewImpl::mousePressEvent(QMouseEvent* event)
{
    TreeWidget::mousePressEvent(event);

    if(event->button() == Qt::RightButton){
        popupMenu->popup(event->globalPos());
    }
}


void ItemTreeViewImpl::keyPressEvent(QKeyEvent* event)
{
    bool processed = true;

    switch(event->key()){
    case Qt::Key_Escape:
        self->clearSelection();
        break;
    case Qt::Key_Delete:
        cutSelectedItems();
        break;
    default:
        processed = false;
        break;
    }
        
    if(!processed && (event->modifiers() & Qt::ControlModifier)){
        processed = true;
        switch(event->key()){
        case Qt::Key_A:
            self->selectAllItems();
            break;
        case Qt::Key_R:
            ItemManager::reloadItems(selectedItemList);
            break;
        case Qt::Key_Plus:
        case Qt::Key_Semicolon:
            zoomFontSize(1);
            break;
        case Qt::Key_Minus:
            zoomFontSize(-1);
            break;
        defaut:
            processed = false;
            break;
        }
    }

    if(!processed){
        TreeWidget::keyPressEvent(event);
    }
}


ItvItem* ItemTreeViewImpl::getItvItem(Item* item)
{
    ItvItem* itvItem = 0;
    ItvItemRef* ref = dynamic_cast<ItvItemRef*>(item->customData(0));
    if(ref){
        itvItem = ref->itvItem;
    }
    return itvItem;
}


ItvItem* ItemTreeViewImpl::getOrCreateItvItem(Item* item)
{
    ItvItem* itvItem = getItvItem(item);
    if(!itvItem){
        itvItem = new ItvItem(item, this);
    }
    return itvItem;
}


void ItemTreeViewImpl::onSubTreeAddedOrMoved(Item* item)
{
    isProceccingSlotForRootItemSignals++;
    
    Item* parentItem = item->parentItem();
    if(parentItem){
        if(parentItem == rootItem){
            insertItem(invisibleRootItem(), item, item->nextItem());
        } else {
            ItvItem* parentItvItem = getItvItem(parentItem);
            if(parentItvItem){
                insertItem(parentItvItem, item, item->nextItem());
            }
        }
    }

    isProceccingSlotForRootItemSignals--;
}


void ItemTreeViewImpl::insertItem(QTreeWidgetItem* parentTwItem, Item* item, Item* nextItem)
{
    ItvItem* itvItem = getOrCreateItvItem(item);
    bool inserted = false;
    if(nextItem){
        ItvItem* nextItvItem = getItvItem(nextItem);
        if(nextItvItem){
            int index = parentTwItem->indexOfChild(nextItvItem);
            parentTwItem->insertChild(index, itvItem);
            inserted = true;
        }
    }
    if(!inserted){
        parentTwItem->addChild(itvItem);
    }
        
    if(!parentTwItem->isExpanded()){
        if(!item->isSubItem()){
            parentTwItem->setExpanded(true);
        }
    }

    for(Item* childItem = item->childItem(); childItem; childItem = childItem->nextItem()){
        insertItem(itvItem, childItem, 0);
    }
}


void ItemTreeViewImpl::onSubTreeRemoved(Item* item, bool isMoving)
{
    isProceccingSlotForRootItemSignals++;
    
    ItvItem* itvItem = getItvItem(item);
    if(itvItem){
        QTreeWidgetItem* parentTwItem = itvItem->parent();
        if(parentTwItem){
            parentTwItem->removeChild(itvItem);
        } else {
            takeTopLevelItem(indexOfTopLevelItem(itvItem));
        }
        delete itvItem;
    }

    isProceccingSlotForRootItemSignals--;
}


void ItemTreeViewImpl::dropEvent(QDropEvent* event)
{
    isDropping = true;
    TreeWidget::dropEvent(event);
    isDropping = false;
}
    

void ItemTreeView::onRowsAboutToBeRemoved(const QModelIndex& parent, int start, int end)
{
    if(impl->isProceccingSlotForRootItemSignals == 0){
        impl->onRowsAboutToBeRemoved(parent, start, end);
    }
}


void ItemTreeViewImpl::onRowsAboutToBeRemoved(const QModelIndex& parent, int start, int end)
{
    connectionsFromRootItem.block();

    QTreeWidgetItem* parentTwItem = itemFromIndex(parent);
    if(!parentTwItem){
        parentTwItem = invisibleRootItem();
    }

    for(int i=start; i <= end; ++i){
        ItvItem* itvItem = dynamic_cast<ItvItem*>(parentTwItem->child(i));
        if(itvItem){
            itvItem->isExpandedBeforeRemoving = itvItem->isExpanded();
            if(!isDropping){
                ItemPtr& item = itvItem->item;
                if(!item->isSubItem()){
                    item->detachFromParentItem();
                }
            }
        }
    }
    
    connectionsFromRootItem.unblock();
}


void ItemTreeView::onRowsInserted(const QModelIndex& parent, int start, int end)
{
    if(impl->isProceccingSlotForRootItemSignals == 0){
        impl->onRowsInserted(parent, start, end);
    }
}


void ItemTreeViewImpl::onRowsInserted(const QModelIndex& parent, int start, int end)
{
    connectionsFromRootItem.block();

    QTreeWidgetItem* parentTwItem = itemFromIndex(parent);
    if(!parentTwItem){
        parentTwItem = invisibleRootItem();
    }

    ItvItem* parentItvItem = dynamic_cast<ItvItem*>(parentTwItem);
    Item* parentItem = parentItvItem ? parentItvItem->item.get() : rootItem.get();

    ItemPtr nextItem = 0;
    if(end + 1 < parentTwItem->childCount()){
        ItvItem* nextItvItem = dynamic_cast<ItvItem*>(parentTwItem->child(end + 1));
        if(nextItvItem){
            nextItem = nextItvItem->item;
        }
    }
    
    for(int i=start; i <= end; ++i){
        ItvItem* itvItem = dynamic_cast<ItvItem*>(parentTwItem->child(i));
        if(itvItem){
            ItemPtr& item = itvItem->item;
            if(!item->isSubItem()){
                parentItem->insertChildItem(item, nextItem, true);
            }
            if(itvItem->isExpandedBeforeRemoving){
                itvItem->setExpanded(true);
            }
        }
    }
    
    connectionsFromRootItem.unblock();
}


void ItemTreeViewImpl::onTreeChanged()
{

}


void ItemTreeViewImpl::onItemAssigned(Item* assigned, Item* srcItem)
{
    ItvItem* itvItem = getItvItem(assigned);
    ItvItem* srcItvItem = getItvItem(srcItem);
    if(itvItem && srcItvItem){
        itvItem->setCheckState(1, srcItvItem->checkState(1));
        QModelIndex index = indexFromItem(itvItem);
        QModelIndex srcIndex = indexFromItem(srcItvItem);
        selectionModel()->select(
            index, srcItvItem->isSelected() ? QItemSelectionModel::Select : QItemSelectionModel::Deselect);
    }
}


bool ItemTreeViewImpl::dropMimeData(QTreeWidgetItem* parent, int index, const QMimeData* data, Qt::DropAction action)
{
    return TreeWidget::dropMimeData(parent, index, data, action);
}


void ItemTreeViewImpl::onSelectionChanged()
{
    selectedItemList.clear();

    QList<QTreeWidgetItem*> selected = selectedItems();
    for(int i=0; i < selected.size(); ++i){
        ItvItem* itvItem = dynamic_cast<ItvItem*>(selected[i]);
        if(itvItem){
            selectedItemList.push_back(itvItem->item.get());
        }
    }

    sigSelectionChanged(selectedItemList);
    sigSelectionOrTreeChanged(selectedItemList);
}


bool ItemTreeView::isItemSelected(ItemPtr item)
{
    return impl->isItemSelected(item.get());
}


bool ItemTreeViewImpl::isItemSelected(Item* item)
{
    ItvItem* itvItem = getItvItem(item);
    if(itvItem){
        return itvItem->isSelected();
    }
    return false;
}


bool ItemTreeView::selectItem(ItemPtr item, bool select)
{
    return impl->selectItem(item.get(), select);
}


bool ItemTreeViewImpl::selectItem(Item* item, bool select)
{
    ItvItem* itvItem = getItvItem(item);
    if(itvItem){
        QModelIndex index = indexFromItem(itvItem);
        selectionModel()->select(index, (select ? QItemSelectionModel::Select : QItemSelectionModel::Deselect));
        return select;
    }
    return false;
}


void ItemTreeView::selectAllItems()
{
    impl->selectAll();
}


void ItemTreeView::clearSelection()
{
    impl->selectionModel()->clearSelection();
}


bool ItemTreeView::isItemChecked(ItemPtr item)
{
    return impl->isItemChecked(item.get());
}


bool ItemTreeViewImpl::isItemChecked(Item* item)
{
    ItvItem* itvItem = getItvItem(item);
    if(itvItem){
        return (itvItem->checkState(1) == Qt::Checked);
    }
    return false;
}
    

bool ItemTreeView::checkItem(ItemPtr item, bool checked)
{
    return impl->checkItem(item.get(), checked);
}


bool ItemTreeViewImpl::checkItem(Item* item, bool checked)
{
    ItvItem* itvItem = getItvItem(item);
    if(itvItem){
        itvItem->setCheckState(1, (checked ? Qt::Checked : Qt::Unchecked));
        return checked;
    }
    return false;
}


SignalProxy< boost::signal<void(const ItemList<>&)> > ItemTreeView::sigSelectionChanged()
{
    return impl->sigSelectionChanged;
}


SignalProxy< boost::signal<void(const ItemList<>&)> > ItemTreeView::sigSelectionOrTreeChanged()
{
    return impl->sigSelectionOrTreeChanged;
}


SignalProxy< boost::signal<void(Item* item, bool isChecked)> > ItemTreeView::sigCheckToggled()
{
    return impl->sigCheckToggled;
}


SignalProxy< boost::signal<void(bool isChecked)> > ItemTreeView::sigCheckToggled(Item* targetItem)
{
    return impl->getOrCreateItvItem(targetItem)->sigCheckToggled;
}


ItemList<>& ItemTreeView::allSelectedItems()
{
    return impl->selectedItemList;
}


void ItemTreeView::extractSelectedItemsOfSubTree(ItemPtr topItem, ItemList<>& io_items)
{
    topItem->traverse(bind(&ItemTreeViewImpl::extractSelectedItemsOfSubTreeTraverse, impl, _1, &io_items));
}


void ItemTreeViewImpl::extractSelectedItemsOfSubTreeTraverse(Item* item, ItemList<>* io_items)
{
    ItvItem* itvItem = getItvItem(item);
    if(itvItem && itvItem->isSelected()){
        io_items->push_back(item);
    }
}


ItemList<>& ItemTreeView::allCheckedItems()
{
    return impl->checkedItems();
}


ItemList<>& ItemTreeViewImpl::checkedItems()
{
    if(needToUpdateCheckedItemList){
        checkedItemList.clear();
        extractCheckedItems(invisibleRootItem());
        needToUpdateCheckedItemList = false;
    }
    return checkedItemList;
}


void ItemTreeViewImpl::extractCheckedItems(QTreeWidgetItem* twItem)
{
    ItvItem* itvItem = dynamic_cast<ItvItem*>(twItem);
    if(itvItem){
        if(itvItem->checkState(1) == Qt::Checked){
            checkedItemList.push_back(itvItem->item.get());
        }
    }
    int n = twItem->childCount();
    for(int i=0; i < n; ++i){
        extractCheckedItems(twItem->child(i));
    }
}


void ItemTreeViewImpl::forEachTopItems(const ItemList<>& items, function<void(Item*)> callback)
{
    set<Item*> itemSet;
    for(size_t i=0; i < items.size(); ++i){
        itemSet.insert(items.get(i));
    }

    for(size_t i=0; i < items.size(); ++i){
        Item* item = items.get(i);
        bool isChild = false;
        Item* parentItem = item->parentItem();
        while(parentItem){
            set<Item*>::iterator p = itemSet.find(parentItem);
            if(p != itemSet.end()){
                isChild = true;
                break;
            }
            parentItem = parentItem->parentItem();
        }
        if(!isChild){
            callback(item);
        }
    }
}


void ItemTreeViewImpl::cutSelectedItems()
{
    copiedItemList.clear();
    ItemList<> selected = selectedItemList;
    forEachTopItems(selected, bind(&ItemTreeViewImpl::moveCutItemsToCopiedItemList, this, _1));
}


void ItemTreeView::cutSelectedItems()
{
    impl->cutSelectedItems();
}


void ItemTreeViewImpl::copySelectedItems()
{
    copiedItemList.clear();

    set<Item*> items;
    for(size_t i=0; i < selectedItemList.size(); ++i){
        items.insert(selectedItemList.get(i));
    }
    
    for(size_t i=0; i < selectedItemList.size(); ++i){
        Item* item = selectedItemList.get(i);
        bool isChild = false;
        Item* parentItem = item->parentItem();
        while(parentItem){
            set<Item*>::iterator p = items.find(parentItem);
            if(p != items.end()){
                isChild = true;
                break;
            }
            parentItem = parentItem->parentItem();
        }
        if(!isChild){
            ItemPtr duplicated = item->duplicate();
            copiedItemList.push_back(duplicated);
            copySelectedItemsSub(item, duplicated, items);
        }
    }
}


void ItemTreeViewImpl::copySelectedItemsSub(Item* item, ItemPtr& duplicated, set<Item*>& items)
{
    for(Item* childItem = item->childItem(); childItem; childItem = childItem->nextItem()){
        set<Item*>::iterator p = items.find(childItem);
        if(p != items.end()){
            ItemPtr duplicatedChild;
            if(childItem->isSubItem()){
                duplicatedChild = duplicated->findItem(childItem->name());
            } else {
                duplicatedChild = childItem->duplicate();
                duplicated->addChildItem(duplicatedChild);
            }
            if(duplicatedChild){
                copySelectedItemsSub(childItem, duplicatedChild, items);
            }
        }
    }
}


void ItemTreeViewImpl::copySelectedItemsWithChildren()
{
    copiedItemList.clear();
    forEachTopItems(selectedItemList, bind(&ItemTreeViewImpl::addCopiedItemToCopiedItemList, this, _1));
}


void ItemTreeViewImpl::addCopiedItemToCopiedItemList(Item* item)
{
    ItemPtr duplicated = item->duplicateAll();
    if(duplicated){
        copiedItemList.push_back(duplicated);
    }
}


void ItemTreeViewImpl::pasteItems()
{
    if(!copiedItemList.empty()){
        ItemPtr parentItem;
        if(selectedItemList.empty()){
            parentItem = rootItem;
        } else if(selectedItemList.size() == 1){
            parentItem = selectedItemList.front();
        }
        if(parentItem){
            for(size_t i=0; i < copiedItemList.size(); ++i){
                ItemPtr org = copiedItemList[i];
                ItemPtr duplicated = org->duplicateAll();
                copiedItemList[i] = duplicated;
                parentItem->addChildItem(org, true); // paste the original items
            }
        }
    }
}


void ItemTreeViewImpl::moveCutItemsToCopiedItemList(Item* item)
{
    if(!item->isSubItem()){
        copiedItemList.push_back(item);
        item->detachFromParentItem();
    }
}


void ItemTreeViewImpl::checkSelectedItems(bool on)
{
    for(size_t i=0; i < selectedItemList.size(); ++i){
        checkItem(selectedItemList[i].get(), on);
    }
}


void ItemTreeViewImpl::toggleSelectedItemChecks()
{
    for(size_t i=0; i < selectedItemList.size(); ++i){
        Item* item = selectedItemList[i].get();
        checkItem(item, !isItemChecked(item));
    }
}


bool ItemTreeView::storeState(Archive& archive)
{
    return impl->storeState(archive);
}


bool ItemTreeViewImpl::storeState(Archive& archive)
{
    storeItemIds(archive, "selected", selectedItemList);
    storeItemIds(archive, "checked", checkedItems());
    storeExpandedItems(archive);
    return true;
}


void ItemTreeViewImpl::storeItemIds(Archive& archive, const char* key, const ItemList<>& items)
{
    ListingPtr idseq = new Listing();
    idseq->setFlowStyle(true);
    for(size_t i=0; i < items.size(); ++i){
        int id = archive.getItemId(items[i]);
        if(id >= 0){
            idseq->append(id);
        }
    }
    if(!idseq->empty()){
        archive.insert(key, idseq);
    }
}


bool ItemTreeView::restoreState(const Archive& archive)
{
    return impl->restoreState(archive);
}


bool ItemTreeViewImpl::restoreState(const Archive& archive)
{
    restoreItemStates(archive, "checked", bind(&ItemTreeView::checkItem, self, _1, true));
    restoreItemStates(archive, "selected", bind(&ItemTreeView::selectItem, self, _1, true));
    restoreExpandedItems(archive);
    return true;
}


void ItemTreeViewImpl::restoreItemStates
(const Archive& archive, const char* key, function<void(ItemPtr)> stateChangeFunc)
{
    const Listing& idseq = *archive.findListing(key);
    if(idseq.isValid()){
        for(int i=0; i < idseq.size(); ++i){
            int id = idseq[i].toInt();
            if(id >= 0){
                ItemPtr item = archive.findItem(id);
                if(item){
                    stateChangeFunc(item);
                }
            }
        }
    }
}


void ItemTreeViewImpl::storeExpandedItems(Archive& archive)
{
    ListingPtr expanded = new Listing();
    expanded->setFlowStyle(true);
    storeExpandedItemsSub(invisibleRootItem(), archive, expanded);
    if(!expanded->empty()){
        archive.insert("expanded", expanded);
    }
}


void ItemTreeViewImpl::storeExpandedItemsSub(QTreeWidgetItem* parentTwItem, Archive& archive, ListingPtr& expanded)
{
    int n = parentTwItem->childCount();
    for(int i=0; i < n; ++i){
        ItvItem* itvItem = dynamic_cast<ItvItem*>(parentTwItem->child(i));
        if(itvItem){
            if(itvItem->isExpanded()){
                Item* item = itvItem->item.get();
                if(!item->isSubItem()){
                    int id = archive.getItemId(item);
                    if(id >= 0){
                        expanded->append(id);
                    }
                } else {
                    int j = 0;
                    Item* p = item;
                    while(p->isSubItem()){
                        ++j;
                        p = p->parentItem();
                    }
                    int id = archive.getItemId(p);
                    if(id >= 0){
                        ListingPtr path = new Listing(j+1);
                        path->setFlowStyle(true);
                        while(item->isSubItem()){
                            path->write(j--, item->name(), DOUBLE_QUOTED);
                            item = item->parentItem();
                        }
                        path->write(0, id);
                        expanded->append(path);
                    }
                }
            }
            if(itvItem->childCount() > 0){
                storeExpandedItemsSub(itvItem, archive, expanded);
            }
        }
    }
}


void ItemTreeViewImpl::restoreExpandedItems(const Archive& archive)
{
    const Listing& expanded = *archive.findListing("expanded");
    if(expanded.isValid()){
        collapseAll();
        for(int i=0; i < expanded.size(); ++i){
            Item* item = 0;
            if(!expanded[i].isListing()){
                item = archive.findItem(expanded[i].toInt());
            } else {
                const Listing& path = *expanded[i].toListing();
                int n = path.size();
                if(n >= 2){
                    item = archive.findItem(path[0].toInt());
                    for(int j=1; item && j < n; ++j){
                        item = item->findItem(path[j].toString());
                    }
                }
            }
            if(item){
                ItvItem* itvItem = getItvItem(item);
                if(itvItem){
                    itvItem->setExpanded(true);
                }
            }
        }
    }
}


void ItemTreeViewImpl::zoomFontSize(int pointSizeDiff)
{
    QFont font = TreeWidget::font();
    font.setPointSize(font.pointSize() + pointSizeDiff);
    setFont(font);
    fontPointSizeDiff += pointSizeDiff;
    AppConfig::archive()->openMapping("ItemTreeView")->write("fontZoom", fontPointSizeDiff);
}
