
/**
   @author Shin'ichiro Nakaoka
*/

#include "VRMLToOSGConverter.h"
#include <osg/Geode>
#include <osg/Geometry>
#include <osg/ShapeDrawable>
#include <osg/Group>
#include <osg/MatrixTransform>
#include <osg/Material>
#include <osg/FrontFace>
#include <osg/LightModel>
#include <osg/Depth>
#include <osgUtil/Optimizer>
#include <osgUtil/SmoothingVisitor>
#include <map>
#include <string>
#include <typeinfo>
#include "gettext.h"

using namespace std;
using namespace boost;
using namespace cnoid;

namespace {
    const bool AVOID_SPECULAR_BUG_OF_OPENGL_DRIVER = true;
}


namespace cnoid {

    class VRMLToOSGConverterImpl
    {
    public:
        VRMLToOSGConverterImpl(std::ostream& os__);
        osg::Node* convert(VRMLNode* vrmlNode);
        osg::Node* convertNode(VRMLNode* vnode);
        osg::Node* convertGroupNode(VRMLGroup* vgroup);
        osg::Group* createTransformNode(VRMLTransform* vt);
        osg::Node* convertShapeNode(VRMLNode* vnode);
        osg::Material* createMaterial(VRMLMaterial* vm);
        osg::Geometry* createGeometryFromIndexedFaceSet(VRMLIndexedFaceSet* vface, float alpha);

        ostream* os_;
        ostream& os() { return *os_; }
        
        bool nonTrianglePolygonsEnabled;
        MFInt32 removedFaceIndices;
        MFInt32 dummyRemovedFaceIndices;
        
        VRMLMaterialPtr defaultMaterial;
        osg::ref_ptr<osg::StateSet> stateSetForTransformWithScaling;
        osg::ref_ptr<osg::StateSet> stateSetForTransformWithUniformScaling;

        typedef map<VRMLNode*, osg::ref_ptr<osg::Material> > VRMLNodeToOsgMaterialMap;
        VRMLNodeToOsgMaterialMap vrmlNodeToOsgMaterialMap;
        
        typedef map<VRMLNode*, osg::Node*> VRMLNodeToOsgNodeMap;
        VRMLNodeToOsgNodeMap vrmlNodeToOsgNodeMap;

        typedef map<VRMLGeometry*, osg::Geometry*> VRMLGeometryToOsgGeometryMap;
        VRMLGeometryToOsgGeometryMap vrmlGeometryToOsgGeometryMap;

        osgUtil::Optimizer optimizer;
    };
}


VRMLToOSGConverter::VRMLToOSGConverter(std::ostream& os)
{
    impl = new VRMLToOSGConverterImpl(os);
}


VRMLToOSGConverterImpl::VRMLToOSGConverterImpl(std::ostream& os__)
{
    os_ = &os__;
    
    nonTrianglePolygonsEnabled = false;
    dummyRemovedFaceIndices.push_back(std::numeric_limits<SFInt32>::max());
    
    defaultMaterial = new VRMLMaterial();

    stateSetForTransformWithScaling = new osg::StateSet();
    stateSetForTransformWithScaling->setMode(GL_NORMALIZE, osg::StateAttribute::ON);

    stateSetForTransformWithUniformScaling = new osg::StateSet();
    stateSetForTransformWithUniformScaling->setMode(GL_RESCALE_NORMAL, osg::StateAttribute::ON);
}


VRMLToOSGConverter::~VRMLToOSGConverter()
{
    delete impl;
}


void VRMLToOSGConverter::enableNonTrianglePolygons(bool on)
{
    impl->nonTrianglePolygonsEnabled = on;
}


osg::Node* VRMLToOSGConverter::convert(VRMLNodePtr vrmlNode)
{
    return impl->convert(vrmlNode.get());
}


osg::Node* VRMLToOSGConverterImpl::convert(VRMLNode* vrmlNode)
{
    osg::Node* node = convertNode(vrmlNode);

    if(node){
        // needed ?
        osg::StateSet* state = node->getOrCreateStateSet();
        osg::LightModel* lightModel = new osg::LightModel;
        lightModel->setTwoSided(true);
        state->setAttributeAndModes(lightModel);

        optimizer.optimize(node);
    }

    return node;
}


osg::Node* VRMLToOSGConverterImpl::convertNode(VRMLNode* vnode)
{
    osg::Node* node = 0;

    VRMLNodeToOsgNodeMap::iterator p = vrmlNodeToOsgNodeMap.find(vnode);
    if(p != vrmlNodeToOsgNodeMap.end()){
        node = p->second;
    } else {

        if(VRMLProtoInstance* protoInstance = dynamic_cast<VRMLProtoInstance*>(vnode)){
            vnode = protoInstance->actualNode.get();
        }

        if(vnode){
            if(VRMLGroup* group = dynamic_cast<VRMLGroup*>(vnode)){
                node = convertGroupNode(group);
            } else if(VRMLShape* shape = dynamic_cast<VRMLShape*>(vnode)){
                node = convertShapeNode(shape);
            }

            if(node){
                node->setName(vnode->defName);
                vrmlNodeToOsgNodeMap.insert(make_pair(vnode, node));
            }
        }
    }
        
    return node;
}


osg::Node* VRMLToOSGConverterImpl::convertGroupNode(VRMLGroup* vgroup)
{
    osg::Group* group;

    if(VRMLTransform* transform = dynamic_cast<VRMLTransform*>(vgroup)){
        group = createTransformNode(transform);
    } else {
        group = new osg::Group;
    }
    
    int num = vgroup->children.size();
    for(int i=0; i < num; i++){
        osg::Node* child = convertNode(vgroup->children[i].get());
        if(child){
            group->addChild(child);
        }
    }
    
    if(group->getNumChildren() == 0){
        group->ref();
        group->unref();
        group = 0;
    }
    
    return group;
}


osg::Group* VRMLToOSGConverterImpl::createTransformNode(VRMLTransform* vt)
{
    osg::MatrixTransform* transform = new osg::MatrixTransform;

    const SFRotation::Vector3& a = vt->rotation.axis();
    osg::Matrix R = osg::Matrix::rotate(vt->rotation.angle(), osg::Vec3d(a[0], a[1], a[2]));
    osg::Matrix T;
    T.setTrans(vt->translation[0], vt->translation[1], vt->translation[2]);
    osg::Matrix C;
    C.setTrans(vt->center[0], vt->center[1], vt->center[2]);
    osg::Matrix Cinv;
    Cinv.invert(C);
    const SFRotation::Vector3& sa = vt->scaleOrientation.axis();
    osg::Vec3d scaleAxis(sa[0], sa[1], sa[2]);
    osg::Matrix SR = osg::Matrix::rotate(vt->scaleOrientation.angle(), scaleAxis);
    osg::Matrix SRinv;
    SRinv.invert(SR);

    const SFVec3f& s = vt->scale;
    osg::Matrix S = osg::Matrix::scale(s[0], s[1], s[2]);
    
    T.preMult(C);
    T.preMult(R);
    T.preMult(SR);
    T.preMult(S);
    T.preMult(SRinv);
    T.preMult(Cinv);
    
    transform->setMatrix(T);

    if((s.array() != 1.0).any()){
        if(s[0] == s[1] && s[1] == s[2]){
            transform->setStateSet(stateSetForTransformWithUniformScaling.get());
        } else {
            transform->setStateSet(stateSetForTransformWithScaling.get());
        }
    }

    return transform;
}


osg::Node* VRMLToOSGConverterImpl::convertShapeNode(VRMLNode* vnode)
{
    VRMLShape* vrmlShape = static_cast<VRMLShape*>(vnode);

    VRMLMaterial* vm = defaultMaterial.get();

    if(vrmlShape->appearance && vrmlShape->appearance->material){
        vm = vrmlShape->appearance->material.get();
    }
    float alpha = 1.0 - vm->transparency;

    osg::Geode* geode = 0;
    osg::Geometry* geometry = 0;
    VRMLGeometry* vrmlGeometry = dynamic_node_cast<VRMLGeometry>(vrmlShape->geometry).get();
    
    if(vrmlGeometry){

        VRMLGeometryToOsgGeometryMap::iterator p = vrmlGeometryToOsgGeometryMap.find(vrmlGeometry);
        if(p != vrmlGeometryToOsgGeometryMap.end()){
            geometry = p->second;
        } else {
            if(VRMLIndexedFaceSet* faceSet = dynamic_cast<VRMLIndexedFaceSet*>(vrmlGeometry)){
                geometry = createGeometryFromIndexedFaceSet(faceSet, alpha);
            }
            if(geometry){
                vrmlGeometryToOsgGeometryMap.insert(make_pair(vrmlGeometry, geometry));
            }
        }
    }
    
    if(geometry){
        geode = new osg::Geode;
        geode->addDrawable(geometry);

        osg::Material* material;

        VRMLNodeToOsgMaterialMap::iterator p = vrmlNodeToOsgMaterialMap.find(vm);
        if(p != vrmlNodeToOsgMaterialMap.end()){
            material = p->second;
        } else {
            material = createMaterial(vm);
            vrmlNodeToOsgMaterialMap.insert(make_pair(vm, material));
        }

        osg::StateSet* stateSet = geode->getOrCreateStateSet();
        stateSet->setAttributeAndModes(material);

        if(alpha < 1.0f){
            // Enable blending, select transparent bin.
            stateSet->setMode(GL_BLEND, osg::StateAttribute::ON);
            stateSet->setRenderingHint(osg::StateSet::TRANSPARENT_BIN);
            
            // Enable depth test so that an opaque polygon will occlude a transparent one behind it.
            stateSet->setMode(GL_DEPTH_TEST, osg::StateAttribute::ON);
            
            // Conversely, disable writing to depth buffer so that
            // a transparent polygon will allow polygons behind it to shine thru.
            // OSG renders transparent polygons after opaque ones.
            osg::Depth* depth = new osg::Depth;
            depth->setWriteMask(false);
            stateSet->setAttributeAndModes(depth, osg::StateAttribute::ON);
            
            // Disable conflicting modes.
            //stateSet->setMode(GL_LIGHTING, osg::StateAttribute::OFF);
        }
    }

    return geode;
}


osg::Material* VRMLToOSGConverterImpl::createMaterial(VRMLMaterial* vm)
{
    osg::Material* material = new osg::Material();

    float alpha = 1.0 - vm->transparency;

    osg::Vec4 diffuse(vm->diffuseColor[0], vm->diffuseColor[1], vm->diffuseColor[2], alpha);
    material->setDiffuse(osg::Material::FRONT_AND_BACK, diffuse);
    float a = vm->ambientIntensity;
    osg::Vec4 ambient(diffuse[0] * a, diffuse[1] * a, diffuse[2] * a, alpha);
    material->setAmbient(osg::Material::FRONT_AND_BACK, ambient);
    osg::Vec4 emission(vm->emissiveColor[0], vm->emissiveColor[1], vm->emissiveColor[2], alpha);
    material->setEmission(osg::Material::FRONT_AND_BACK, emission);

    osg::Material::Face specularFace =
        AVOID_SPECULAR_BUG_OF_OPENGL_DRIVER ? osg::Material::FRONT : osg::Material::FRONT_AND_BACK;
    osg::Vec4 specular(vm->specularColor[0], vm->specularColor[1], vm->specularColor[2], 1.0);
    material->setSpecular(specularFace, specular);
    float shininess = (127.0 * vm->shininess) + 1.0;
    material->setShininess(specularFace, shininess);

    return material;
}


osg::Geometry* VRMLToOSGConverterImpl::createGeometryFromIndexedFaceSet(VRMLIndexedFaceSet* vface, float alpha)
{
    osg::Geometry* geometry = new osg::Geometry;
    
    if(!vface->ccw){
        osg::StateSet* stateSet = geometry->getOrCreateStateSet();
        stateSet->setAttributeAndModes(new osg::FrontFace(osg::FrontFace::CLOCKWISE));
    }

    if(!vface->solid){
        
    }
    
    if(vface->coord){
        osg::Vec3Array* vertices = new osg::Vec3Array;
        MFVec3s& point = vface->coord->point;
        int size = point.size();
        for(int i=0; i < size; i++){
            vertices->push_back(osg::Vec3(point[i][0], point[i][1], point[i][2]));
        }
        geometry->setVertexArray(vertices);
    }

    removedFaceIndices.clear();

    if(nonTrianglePolygonsEnabled){
        geometry->addPrimitiveSet(new osg::DrawArrayLengths(osg::PrimitiveSet::POLYGON));
        osg::IntArray* vertexIndices = new osg::IntArray();
        int numVertex = 0;
        int numIndices = vface->coordIndex.size();
        
        for(int i=0; i < numIndices; i++) {
            int index = vface->coordIndex[i];
            if (index < 0) {
                ((osg::DrawArrayLengths*)geometry->getPrimitiveSet(0))->push_back(numVertex);
                numVertex = 0;
            } else {
                vertexIndices->push_back(index);
                numVertex++;
            }
        }
        geometry->setVertexIndices(vertexIndices);

    } else {
        osg::DrawArrays* drawArrays = new osg::DrawArrays(osg::PrimitiveSet::TRIANGLES);
        osg::IntArray* vertexIndices = new osg::IntArray();

        const MFInt32& coordIndex = vface->coordIndex;
        int numIndices = coordIndex.size();

        int faceIndex = 0;
        int faceVertexIndexBegin = 0;
        int numFaceVertices = 0;
        
        for(int i=0; i < numIndices; ++i){
            int index = coordIndex[i];
            if(index >= 0){
                ++numFaceVertices;
            } else {
                if(numFaceVertices == 3) { // Triangle ?
                    vertexIndices->push_back(coordIndex[faceVertexIndexBegin++]);
                    vertexIndices->push_back(coordIndex[faceVertexIndexBegin++]);
                    vertexIndices->push_back(coordIndex[faceVertexIndexBegin]);
                } else {
                    removedFaceIndices.push_back(faceIndex);
                }
                faceVertexIndexBegin = i + 1; // next index
                numFaceVertices = 0;
                ++faceIndex;
            }
        }
        geometry->setVertexIndices(vertexIndices);
        drawArrays->setCount(vertexIndices->size());
        geometry->addPrimitiveSet(drawArrays);
    }

    if(!removedFaceIndices.empty()){
        os() << str(fmt(_("There are %2% non-triangle polygon(s) in a IndexedFaceSet node%1%."))
                    % (vface->defName.empty() ? "" : (string(" ") + vface->defName))
                    % removedFaceIndices.size())
             << endl;
    }

    removedFaceIndices.push_back(std::numeric_limits<SFInt32>::max()); // dummy

    if(vface->normal){

        osg::Vec3Array* normals = new osg::Vec3Array;
        MFVec3s& vec = vface->normal->vector;
        int size = vec.size();
        for(int i=0; i < size; i++){
            normals->push_back(osg::Vec3(vec[i][0], vec[i][1], vec[i][2]));
        }
        geometry->setNormalArray(normals);
        
        if(vface->normalIndex.empty()){
            geometry->setNormalIndices(geometry->getVertexIndices());
        } else {

            size_t skipIndex = 0;
            MFInt32& indicesToSkip = vface->normalPerVertex ? dummyRemovedFaceIndices : removedFaceIndices;
            int size = vface->normalIndex.size();
            osg::UIntArray* indices = new osg::UIntArray;
            for(int i=0; i < size; ++i){
                if(i == indicesToSkip[skipIndex]){
                    if(skipIndex < indicesToSkip.size()){
                        ++skipIndex;
                    }
                } else {
                    int index = vface->normalIndex[i];
                    if(index >= 0){
                        indices->push_back(index);
                    }
                }
            }
            geometry->setNormalIndices(indices);
        }
        if(vface->normalPerVertex == true){
            geometry->setNormalBinding(osg::Geometry::BIND_PER_VERTEX);
        } else {
            geometry->setNormalBinding(osg::Geometry::BIND_PER_PRIMITIVE);
        }
    }
    
    if(vface->color){
        MFColor& c = vface->color->color;
        if(!c.empty()){
            osg::Vec4Array* colors = new osg::Vec4Array;
            int size = c.size();
            for(int i=0; i < size; i++){
                colors->push_back(osg::Vec4(c[i][0], c[i][1], c[i][2], alpha));
            }
            geometry->setColorArray(colors);
            
            if(vface->colorIndex.empty()){
                geometry->setColorIndices(geometry->getVertexIndices());
            } else {
                size_t skipIndex = 0;
                MFInt32& indicesToSkip = vface->colorPerVertex ? dummyRemovedFaceIndices : removedFaceIndices;
                int size = vface->colorIndex.size();
                osg::UIntArray* indices = new osg::UIntArray;
                
                // osg::TemplateIndexArray <unsigned int, osg::Array::UIntArrayType,4,4> *indices;
                // indices = new osg::TemplateIndexArray<unsigned int, osg::Array::UIntArrayType,4,4>;
                
                for(int i=0; i < size; i++){
                    if(i == indicesToSkip[skipIndex]){
                        if(skipIndex < indicesToSkip.size()){
                            ++skipIndex;
                        }
                    } else {
                        int index = vface->colorIndex[i];
                        if(index >= 0){
                            indices->push_back(index);
                        }
                    }
                }
                geometry->setColorIndices(indices);
            }
            
            if(vface->colorPerVertex == true){
                geometry->setColorBinding(osg::Geometry::BIND_PER_VERTEX);
            } else {
                geometry->setColorBinding(osg::Geometry::BIND_PER_PRIMITIVE);
            }
        }
    }
    
    if(!vface->normal){
        osgUtil::SmoothingVisitor::smooth(*geometry);
    }
    
    return geometry;
}
