/**
   \file
   \author Shin'ichiro Nakaoka
*/

#ifndef CNOID_BODYPLUGIN_BODY_ITEM_H_INCLUDED
#define CNOID_BODYPLUGIN_BODY_ITEM_H_INCLUDED

#include <cnoid/Item>
#include <cnoid/Body>
#include <cnoid/ColdetLinkPair>
#include <cnoid/SceneProvider>
#include <boost/dynamic_bitset.hpp>
#include "exportdecl.h"

namespace cnoid {

    class BodyState;
    class ModelNodeSet;
    typedef boost::shared_ptr<ModelNodeSet> ModelNodeSetPtr;
    class BodyItem;
    typedef boost::intrusive_ptr<BodyItem> BodyItemPtr;
    class InverseKinematics;
    typedef boost::shared_ptr<InverseKinematics> InverseKinematicsPtr;
    class PinDragIK;
    typedef boost::shared_ptr<PinDragIK> PinDragIKptr;
    class PenetrationBlocker;
    typedef boost::shared_ptr<PenetrationBlocker> PenetrationBlockerPtr;
    class BodyItemImpl;


    class CNOID_EXPORT BodyItem : public Item, public SceneProvider
    {
      public:

        static void initialize(ExtensionManager* ext);
        
        BodyItem();
        BodyItem(const BodyItem& org);
        virtual ~BodyItem();

        bool loadModelFile(const std::string& filename);
            
        virtual void setName(const std::string& name);

        const BodyPtr& body() const;

        //! Deprecated. Please use Link::shape
        const ModelNodeSetPtr& modelNodeSet() const;
        
        const std::string& errorMessage() const;

        enum PresetPoseID { INITIAL_POSE, STANDARD_POSE };

        void moveToOrigin();

        void setPresetPose(PresetPoseID id);

        Link* currentBaseLink() const;
        void setCurrentBaseLink(Link* link);

        void calcForwardKinematics(bool calcVelocity = false, bool calcAcceleration = false);

        void copyKinematicState();
        void pasteKinematicState();

        void storeKinematicState(BodyState& state);
        bool restoreKinematicState(const BodyState& state);

        void storeInitialState();
        void restoreInitialState();

        void getInitialState(BodyState& out_state);

        // for undo, redo operations
        void beginKinematicStateEdit();
        void acceptKinematicStateEdit();
        bool undoKinematicState();
        bool redoKinematicState();

        PinDragIKptr pinDragIK();
        InverseKinematicsPtr getCurrentIK(Link* targetLink);
        InverseKinematicsPtr getDefaultIK(Link* targetLink);
        PenetrationBlockerPtr createPenetrationBlocker(Link* link, bool excludeSelfCollisions = false);

        /**
           @if jp
           ロボットの関節角、関節角速度、root位置・姿勢などの「運動学的」状態に変更が生じたときに
           発行されるシグナル。
           Item::sigUpdated() はモデル自体が変わった場合とし、そちらとは区別して使う。
           @endif
        */
        SignalProxy< boost::signal<void()> > sigKinematicStateChanged();

        void notifyKinematicStateChange(
            bool requestFK = false, bool requestVelFK = false, bool requestAccFK = false);
            
        void notifyKinematicStateChange(
            boost::signals::connection& connectionToBlock,
            bool requestFK = false, bool requestVelFK = false, bool requestAccFK = false);

        SignalProxy< boost::signal<void()> > sigKinematicStateEdited();

        void updateColdetModelPositions(bool force = false);

        void enableSelfCollisionDetection(bool on);

        bool isSelfCollisionDetectionEnabled() const;
        bool updateSelfCollisions(bool force = false);
        void clearSelfCollisions();

        std::vector<ColdetLinkPairPtr>& selfColdetPairs();
        boost::dynamic_bitset<>& selfCollisionLinkBitSet();

        SignalProxy< boost::signal<void()> > sigSelfCollisionsUpdated();
        SignalProxy< boost::signal<void()> > sigSelfCollisionLinkSetChanged();

        std::vector<ColdetLinkPairPtr>& worldColdetPairsOfLink(int linkIndex);
        const std::vector<ColdetLinkPairPtr>& worldColdetPairsOfLink(int linkIndex) const;
        boost::dynamic_bitset<>& worldCollisionLinkBitSet();

        SignalProxy< boost::signal<void()> > sigWorldCollisionsUpdated();
        SignalProxy< boost::signal<void()> > sigWorldCollisionLinkSetChanged();
        
        void notifyWorldCollisionLinkSetChange();
        void notifyWorldCollisionUpdate();

        const Vector3& centerOfMass();

        bool doLegIkToMoveCm(const Vector3& c, bool onlyProjectionToFloor = false);

        const Vector3& zmp() const;
        void setZmp(const Vector3& zmp);
        void editZmp(const Vector3& zmp);

        enum PositionType { CM_PROJECTION, HOME_COP, RIGHT_HOME_COP, LEFT_HOME_COP, ZERO_MOMENT_POINT };
            
        boost::optional<Vector3> getParticularPosition(PositionType posType);

        bool setStance(double width);
            
        virtual SgNode* scene();

      protected:
        virtual ItemPtr doDuplicate() const;
        virtual void doAssign(Item* item);
        virtual void doPutProperties(PutPropertyFunction& putProperty);
        virtual bool store(Archive& archive);
        virtual bool restore(const Archive& archive);
            
      private:
        BodyItemImpl* impl;
        friend class BodyItemImpl;
    };
}

#endif
