/**
   @file
   @author Shin'ichiro Nakaoka
*/

#include "WorldItem.h"
#include "KinematicsBar.h"
#include <cnoid/MessageView>
#include <cnoid/ItemManager>
#include <cnoid/RootItem>
#include <cnoid/PutPropertyFunction>
#include <cnoid/YAMLReader>
#include <cnoid/Archive>
#include <cnoid/ConnectionSet>
#include <cnoid/LazyCaller>
#include <cnoid/GLSceneRenderer>
#include <boost/bind.hpp>
#include <boost/dynamic_bitset.hpp>
#include <boost/make_shared.hpp>
#ifdef _WIN32
#include <Windows.h>
#endif
#ifdef __APPLE__
#include <OpenGL/gl.h>
#else
#include <GL/gl.h>
#endif
#include <vector>
#include <algorithm>
#include "gettext.h"

using namespace std;
using namespace boost;
using namespace cnoid;

namespace {

    const bool TRACE_FUNCTIONS = false;

    struct BodyItemInfo {
        BodyItemInfo(int numLinks)
            : worldCollisionLinkBitSet(numLinks) {
            kinematicStateChanged = false;
            worldCollisionLinkSetChanged = false;
            numValidPairs = 0;
        }
        typedef dynamic_bitset<> updateFlag;
        bool kinematicStateChanged;
        bool worldCollisionLinkSetChanged;
        boost::dynamic_bitset<> worldCollisionLinkBitSet;
        int numValidPairs;
    };
    typedef map<BodyItem*, BodyItemInfo> BodyItemInfoMap;

    
    class ColdetLinkPairEx : public ColdetLinkPair
    {
    public:
        BodyItemInfo* info[2];
        
        ColdetLinkPairEx(
            BodyItemInfo* p1, const BodyPtr& body1, Link* link1,
            BodyItemInfo* p2, const BodyPtr& body2, Link* link2)
            : ColdetLinkPair(body1, link1, body2, link2) {
            info[0] = p1;
            info[1] = p2;
        }
    };
    typedef boost::shared_ptr<ColdetLinkPairEx> ColdetLinkPairExPtr;


    class SceneCollision : public SgCustomGLNode
    {
    public:
        const vector<ColdetLinkPairPtr>& pairs;
        signals::connection collisionUpdatedConnection;
        SgBoundingBox bbox;
        
        SceneCollision(WorldItem* worldItem) : pairs(worldItem->coldetLinkPairs()) {

        }

        virtual void render(GLSceneRenderer& renderer) {

            if(!renderer.property()->get("collision", false)){
                return;
            }

            renderer.enableLighting(false);
            renderer.setColor(Vector4f(0.0f, 1.0f, 0.0f, 1.0f));

            glBegin(GL_LINES);

            for(size_t i=0; i < pairs.size(); ++i){
                const std::vector<collision_data>& cols = pairs[i]->collisions();
                for(size_t j=0; j < cols.size(); ++j){
                    const collision_data& col = cols[j];
                    const Vector3 n = 50.0 * col.depth * col.n_vector;
                    for(int k=0; k < col.num_of_i_points; ++k){
                        if(col.i_point_new[k]){
                            const Vector3& p = col.i_points[k];
                            glVertex3dv(p.data());
                            glVertex3dv(Vector3(p + n).data());
                        }
                    }
                }
            }
        
            glEnd();

            renderer.enableLighting(true);
        }

        virtual const SgBoundingBox& boundingBox() const {
            return bbox;
        }
    };

    typedef intrusive_ptr<SceneCollision> SceneCollisionPtr;
}



namespace cnoid {

    class WorldItemImpl
    {
    public:
        WorldItemImpl(WorldItem* self);
        WorldItemImpl(WorldItem* self, const WorldItemImpl& org);
        ~WorldItemImpl();
            
        WorldItem* self;
        ostream& os;

        ItemList<BodyItem> bodyItems;

        signals::connection sigItemTreeChangedConnection;
        ConnectionSet sigKinematicStateChangedConnections;

        bool isCollisionDetectionEnabled;
        LazyCaller updateCollisionsCaller;
        KinematicsBar* kinematicsBar;

        BodyItemInfoMap bodyItemInfoMap;

        std::vector<ColdetLinkPairPtr> coldetPairs;

        boost::signal<void()> sigColdetPairsUpdated;
        boost::signal<void()> sigCollisionsUpdated;

        SceneCollisionPtr sceneCollision;

        void init();
        void enableCollisionDetection(bool on);
        void clearColdetLinkPairs();
        void updateColdetLinkPairs(bool forceUpdate);
        void onBodyKinematicStateChanged(BodyItem* bodyItem);
        void updateCollisions(bool forceUpdate);
    };
}


void cnoid::initializeWorldItem(ExtensionManager& ext)
{
    ext.itemManager().registerClass<WorldItem>(N_("WorldItem"));
    ext.itemManager().addCreationPanel<WorldItem>();
}


WorldItem::WorldItem()
{
    impl = new WorldItemImpl(this);
}


WorldItemImpl::WorldItemImpl(WorldItem* self)
    : self(self),
      os(MessageView::mainInstance()->cout()),
      updateCollisionsCaller(bind(&WorldItemImpl::updateCollisions, this, false), IDLE_PRIORITY_HIGH)
{
    isCollisionDetectionEnabled = false;
    init();
}


WorldItem::WorldItem(const WorldItem& org)
    : Item(org)
{
    impl = new WorldItemImpl(this, *org.impl);
}


WorldItemImpl::WorldItemImpl(WorldItem* self, const WorldItemImpl& org)
    : self(self),
      os(org.os),
      updateCollisionsCaller(bind(&WorldItemImpl::updateCollisions, this, false), IDLE_PRIORITY_HIGH)
{
    isCollisionDetectionEnabled = org.isCollisionDetectionEnabled;
    init();
}


void WorldItemImpl::init()
{
    kinematicsBar = KinematicsBar::instance();
}
    
WorldItem::~WorldItem()
{
    delete impl;
}


WorldItemImpl::~WorldItemImpl()
{
    sigKinematicStateChangedConnections.disconnect();
    sigItemTreeChangedConnection.disconnect();
}


const std::vector<ColdetLinkPairPtr>& WorldItem::coldetLinkPairs() const
{
    return impl->coldetPairs;
}


void WorldItem::enableCollisionDetection(bool on)
{
    impl->enableCollisionDetection(on);
}


void WorldItemImpl::enableCollisionDetection(bool on)
{
    if(TRACE_FUNCTIONS){
        os << "WorldItemImpl::enableCollisionDetection(" << on << ")" << endl;
    }

    bool changed = false;
    
    if(isCollisionDetectionEnabled && !on){
        clearColdetLinkPairs();
        sigItemTreeChangedConnection.disconnect();
        isCollisionDetectionEnabled = false;
        changed = true;
        
    } else if(!isCollisionDetectionEnabled && on){
        isCollisionDetectionEnabled = true;
        updateColdetLinkPairs(true);
        sigItemTreeChangedConnection =
            RootItem::mainInstance()->sigTreeChanged().connect(
                bind(&WorldItemImpl::updateColdetLinkPairs, this, false));
        changed = true;
    }

    if(changed){
        sigColdetPairsUpdated();
        self->notifyUpdate();
        sigCollisionsUpdated();
    }
}


bool WorldItem::isCollisionDetectionEnabled()
{
    return impl->isCollisionDetectionEnabled;
}


void WorldItemImpl::clearColdetLinkPairs()
{
    if(TRACE_FUNCTIONS){
        os << "WorldItemImpl::clearColdetLinkPairs()" << endl;
    }
    
    coldetPairs.clear();
    sigKinematicStateChangedConnections.disconnect();
    bodyItemInfoMap.clear();

    for(size_t i=0; i < bodyItems.size(); ++i){
        BodyItem* item = bodyItems.get(i);
        size_t n = item->body()->numLinks();
        for(size_t j=0; j < n; ++j){
            item->worldColdetPairsOfLink(j).clear();
        }
        bodyItemInfoMap.insert(make_pair(item, BodyItemInfo(n)));
    }
}    


void WorldItemImpl::updateColdetLinkPairs(bool forceUpdate)
{
    if(TRACE_FUNCTIONS){
        os << "WorldItemImpl::updateColdetLinkPairs()" << endl;
    }

    if(!forceUpdate){
        ItemList<BodyItem> prevBodyItems = bodyItems;
        bodyItems.extractChildItems(self);
        if(bodyItems == prevBodyItems){
            return;
        }
    } else {
        bodyItems.extractChildItems(self);
    }
        
    clearColdetLinkPairs();

    for(BodyItemInfoMap::iterator p1 = bodyItemInfoMap.begin(); p1 != bodyItemInfoMap.end(); ++p1){

        BodyItem* bodyItem1 = p1->first;
        BodyItemInfo& info1 = p1->second;
        const BodyPtr& body1 = bodyItem1->body();
        
        for(BodyItemInfoMap::iterator p2 = p1; p2 != bodyItemInfoMap.end(); ++p2){

            BodyItem* bodyItem2 = p2->first;
            BodyItemInfo& info2 = p2->second;
            const BodyPtr& body2 = bodyItem2->body();
                    
            for(int l1 = 0; l1 < body1->numLinks(); ++l1){
                int l2;
                if(p1 == p2){
                    continue;
                    //l2 = l1 + 1;
                } else {
                    l2 = 0;
                }
                Link* link1 = body1->link(l1);
                std::vector<ColdetLinkPairPtr>& pairs1 = bodyItem1->worldColdetPairsOfLink(l1);
                
                for( ; l2 < body2->numLinks(); ++l2){
                    Link* link2 = body2->link(l2);
                    //if(link1 != link2){
                    if(link1->parent() != link2 && link2->parent() != link1){
                        std::vector<ColdetLinkPairPtr>& pairs2 = bodyItem2->worldColdetPairsOfLink(l2);
                        ColdetLinkPairExPtr linkPair = make_shared<ColdetLinkPairEx>(&info1, body1, link1, &info2, body2, link2);
                        coldetPairs.push_back(linkPair);
                        pairs1.push_back(linkPair);
                        pairs2.push_back(linkPair);
                        info1.numValidPairs++;
                        info2.numValidPairs++;
                    }
                }
            }
        }
    }

    for(BodyItemInfoMap::iterator p = bodyItemInfoMap.begin(); p != bodyItemInfoMap.end(); ++p){
        BodyItemInfo& info = p->second;
        if(info.numValidPairs > 0){
            BodyItem* bodyItem = p->first;
            sigKinematicStateChangedConnections.add(
                bodyItem->sigKinematicStateChanged().connect(
                    bind(&WorldItemImpl::onBodyKinematicStateChanged, this, bodyItem)));
        }
    }

    if(isCollisionDetectionEnabled){
        updateCollisions(true);
    }
}


void WorldItemImpl::onBodyKinematicStateChanged(BodyItem* bodyItem)
{
    if(TRACE_FUNCTIONS){
        os << "WorldItemImpl::onBodyKinematicStateChanged()" << endl;
    }
    
    BodyItemInfoMap::iterator p = bodyItemInfoMap.find(bodyItem);
    if(p != bodyItemInfoMap.end()){
        BodyItemInfo& info = p->second;
        info.kinematicStateChanged = true;
        updateCollisionsCaller.setPriority(kinematicsBar->collisionDetectionPriority());
        updateCollisionsCaller.request();
    }
}


void WorldItem::updateCollisions()
{
    impl->updateCollisions(true);
}


void WorldItemImpl::updateCollisions(bool forceUpdate)
{
    if(TRACE_FUNCTIONS){
        os << "WorldItemImpl::updateCollisions()" << endl;
    }

    for(BodyItemInfoMap::iterator p = bodyItemInfoMap.begin(); p != bodyItemInfoMap.end(); ++p){
        BodyItem* bodyItem = p->first;
        BodyItemInfo& info = p->second;
        bodyItem->updateColdetModelPositions(forceUpdate);
        // this should not called here
        //info.worldCollisionLinkSetChanged = bodyItem->updateSelfCollisions(forceUpdate); 
        if(forceUpdate){
            info.kinematicStateChanged = true;
        }
    }

    for(size_t i=0; i < coldetPairs.size(); ++i){

        ColdetLinkPairEx* linkPair = static_cast<ColdetLinkPairEx*>(coldetPairs[i].get());
        BodyItemInfo* info1 = linkPair->info[0];
        BodyItemInfo* info2 = linkPair->info[1];
        
        if(info1->kinematicStateChanged || info2->kinematicStateChanged){
            
            bool prevEmpty = linkPair->collisions().empty();
            //linkPair->updatePositions();
            bool empty = linkPair->detectCollisions().empty();
            if(prevEmpty != empty){
                info1->worldCollisionLinkSetChanged = true;
                info2->worldCollisionLinkSetChanged = true;
            }
            if(!empty){
                info1->worldCollisionLinkBitSet.set(linkPair->link(0)->index());
                info2->worldCollisionLinkBitSet.set(linkPair->link(1)->index());
            }
        }
    }
    
    for(BodyItemInfoMap::iterator p = bodyItemInfoMap.begin(); p != bodyItemInfoMap.end(); ++p){
        
        BodyItem* bodyItem = p->first;
        BodyItemInfo& info = p->second;
        info.kinematicStateChanged = false;

        bodyItem->worldCollisionLinkBitSet() = (info.worldCollisionLinkBitSet | bodyItem->selfCollisionLinkBitSet());
        
        if(info.worldCollisionLinkSetChanged){
            bodyItem->notifyWorldCollisionLinkSetChange();
        }
        bodyItem->notifyWorldCollisionUpdate();
        
        info.worldCollisionLinkSetChanged = false;
        info.worldCollisionLinkBitSet.reset();
    }
        
    sigCollisionsUpdated();
}


SignalProxy< boost::signal<void()> > WorldItem::sigColdetPairsUpdated()
{
    return impl->sigColdetPairsUpdated;
}


SignalProxy< boost::signal<void()> > WorldItem::sigCollisionsUpdated()
{
    return impl->sigCollisionsUpdated;
}


SgNode* WorldItem::scene()
{
    if(!impl->sceneCollision){
        impl->sceneCollision = new SceneCollision(this);
    }
    return impl->sceneCollision.get();
}


ItemPtr WorldItem::doDuplicate() const
{
    return new WorldItem(*this);
}


void WorldItem::doPutProperties(PutPropertyFunction& putProperty)
{
    putProperty(_("Collision detection"), isCollisionDetectionEnabled(),
                bind(&WorldItem::enableCollisionDetection, this, _1), true);
}


bool WorldItem::store(Archive& archive)
{
    archive.write("collisionDetection", isCollisionDetectionEnabled());
    return true;
}


bool WorldItem::restore(const Archive& archive)
{
    if(archive.get("collisionDetection", false)){
        archive.addPostProcess(bind(&WorldItemImpl::enableCollisionDetection, impl, true));
    }
    return true;
}
